//colorLlineSensor_v2
//v21	20231105	動作テスト

#include <Wire.h>
#include <EEPROM.h>

//****	定数宣言  ******************************************

//------   ポート配置    -------------------------
#define SW 8									//白補正SWのポート
#define LED_R 7//5						//LEDのポート 各色
#define LED_G 6
#define LED_B 5//7
#define PT1 A0								//フォトトランジスタ入力ポート
#define PT2 A1
#define PT3 A2
#define PT4 A3
#define PT5 A6

//------   基本数値    ---------------------------
#define	NUM_PH	5      				// センサの数
#define NUM_CNT 4							//サンプリング回数
#define RED			0							//データ配列の番号
#define GREEN		1
#define BLUE		2
#define DARK		3

//****	グローバル変数  ************************************

static int sCnt;												//サンプリングカウンタ

//センサの読み取り値[サンプル番号][センサ番号][色：0=R,1=G,2=B,3=消灯] 平均化前
static word adIn[NUM_CNT][NUM_PH][4];
static word adData[NUM_PH][4];					//読み取り値 平均補正後

static word wK[NUM_PH][3];  						// 白補正の係数
static word cData[NUM_PH][3];    				// 補正後の値
static word cData0[NUM_PH];        			// 白
static int i1, i2;											//一時変数
static long hue[NUM_PH], hue255_i1;			//色相データ
static word hue255[3];									//色相　8bit
static word rgb255[NUM_PH][3];					//輝度
static word add1;												//累計バッファ
//static word addC[NUM_PH], add1;					//累計バッファ
//static long hi, f;

word sndW[12];        									// 送信データ（wordサイズ）
byte sBuf[24];          								// 送信データのバッファ
byte rBuf[10];          								// 受信データのバッファ
byte numCmd;    												// コマンドのバイト数（2バイト目以降）
byte command;            								// マスタから受け取ったコマンド
int numRep;        											// マスタに返すバイト数


//****	セットアップ  **************************************
void setup() {
  int i, j, k;                          // ループのカウンタ
//  byte wh, wl;

  sCnt = 0;

  Wire.begin(0x18);											// アドレスを設定
  Wire.onReceive(i2cReceiveEvent);      // 割り込み関数の登録
  Wire.onRequest(i2cSendEvent);        	// 割り込み関数の登録

//  Serial.begin(57600);

  pinMode(LED_R, OUTPUT);								//LEDポート設定
  pinMode(LED_G, OUTPUT);
  pinMode(LED_B, OUTPUT);
  pinMode(PT1, INPUT);									//フォトトランジスタポート設定
  pinMode(PT2, INPUT);
  pinMode(PT3, INPUT);
  pinMode(PT4, INPUT);
  pinMode(PT5, INPUT);
  pinMode(SW, INPUT_PULLUP);						//SWポート設定

	Read_EEP();														// 補正値をEEPROMから読み出す
}
//****	メインループ  **************************************
void loop() {

	Read_Color();												//LED点灯　色反射読取り動作
	Data_Correct();											//データ累計と補正
	Hue_Correct();											//色相算出

  if (digitalRead(SW) == LOW) {				//白補正押している間動作
		adjust_White();										//白補正動作
	}
}

//****	測定と補正  ****************************************

void Read_Color()												//LED点灯　色反射読取り動作
{

  digitalWrite(LED_R, LOW);								//消灯時データ
  digitalWrite(LED_G, LOW);
  digitalWrite(LED_B, LOW);
  delayMicroseconds(600);
  adIn[sCnt][0][DARK] = analogRead(PT1);
  adIn[sCnt][1][DARK] = analogRead(PT2);
  adIn[sCnt][2][DARK] = analogRead(PT3);
  adIn[sCnt][3][DARK] = analogRead(PT4);
  adIn[sCnt][4][DARK] = analogRead(PT5);

  digitalWrite(LED_R, HIGH);							//赤データ
  digitalWrite(LED_G, LOW);
  digitalWrite(LED_B, LOW);
  delayMicroseconds(600);
  adIn[sCnt][0][RED] = analogRead(PT1);
  adIn[sCnt][1][RED] = analogRead(PT2);
  adIn[sCnt][2][RED] = analogRead(PT3);
  adIn[sCnt][3][RED] = analogRead(PT4);
  adIn[sCnt][4][RED] = analogRead(PT5);

  digitalWrite(LED_R, LOW);								//緑データ
  digitalWrite(LED_G, HIGH);
  digitalWrite(LED_B, LOW);
  delayMicroseconds(600);
  adIn[sCnt][0][GREEN] = analogRead(PT1);
  adIn[sCnt][1][GREEN] = analogRead(PT2);
  adIn[sCnt][2][GREEN] = analogRead(PT3);
  adIn[sCnt][3][GREEN] = analogRead(PT4);
  adIn[sCnt][4][GREEN] = analogRead(PT5);

  digitalWrite(LED_R, LOW);								//青データ
  digitalWrite(LED_G, LOW);
  digitalWrite(LED_B, HIGH);
  delayMicroseconds(600);
  adIn[sCnt][0][BLUE] = analogRead(PT1);
  adIn[sCnt][1][BLUE] = analogRead(PT2);
  adIn[sCnt][2][BLUE] = analogRead(PT3);
  adIn[sCnt][3][BLUE] = analogRead(PT4);
  adIn[sCnt][4][BLUE] = analogRead(PT5);

  digitalWrite(LED_R, LOW);								//消灯
  digitalWrite(LED_G, LOW);
  digitalWrite(LED_B, LOW);

  sCnt++;																	//カウンタ動作
  if (sCnt >= NUM_CNT) {									//サンプリング数だけ
    sCnt = 0;
  }
}
//------   データ累計と補正    -------------------
void Data_Correct()											//データ累計と補正
{
  int i, j, k;

  for (j = 0; j < NUM_PH; j++) {									//累計　平均値算出
    for (i = 0; i < 4; i++) {
      adData[j][i] = 0;
      for (k = 0; k < NUM_CNT; k++) {
        adData[j][i] += adIn[k][j][i];
      }
      adData[j][i] = adData[j][i] >> 2;
    }
  }

  for (i = 0; i < NUM_PH; i++) {             			// 消灯時の電圧を引く
    for (j = 0; j < 3; j++) {
      adData[i][j] = adData[i][j] - adData[i][DARK];
    }
  }

  for (i = 0; i < NUM_PH; i++) {
    for (j = 0; j < 3; j++) {
      cData[i][j] = adData[i][j] * wK[i][j];    	// 白補正 16倍のまま
    }
  }

  for (i = 0; i < NUM_PH; i++) {									// RGBを0～255に正規化
    i2 = cData0[i] / 256;													//　輝度計算
    add1 = 0;
    for (j = 0; j < 3; j++) {
      rgb255[i][j] = cData[i][j] / i2;
      add1 += rgb255[i][j];
      if (rgb255[i][j] > 0xff) {
        rgb255[i][j] = 0xff;
      }
    }
    sndW[i + 5] = add1 * 2 / 3;
  }
}
//------   色合い補正    -------------------------
void Hue_Correct()											//色相算出
{
  int i;
  int mx, mn;

  for (i = 0; i < NUM_PH; i++) {
    mx = seek_Max(&(rgb255[i][0]), 3);
    mn = seek_Min(&(rgb255[i][0]), 3);
    switch (mx) {													//一番明るい色に合わせて計算
      case 0:
        hue[i] = (long)((60 * ((long)rgb255[i][1] - (long)rgb255[i][2])
								/ ((long)rgb255[i][mx] - (long)rgb255[i][mn]) + 360) % 360);
        break;
      case 1:
        hue[i] = (long)((60 * ((long)rgb255[i][2] - (long)rgb255[i][0])
								/ ((long)rgb255[i][mx] - (long)rgb255[i][mn])) + 120);
        break;
      case 2:
        hue[i] = (long)((60 * ((long)rgb255[i][0] - (long)rgb255[i][1]) 
								/ ((long)rgb255[i][mx] - (long)rgb255[i][mn])) + 240);
        break;
      default:
        hue[i] = 0;
        break;
    }
    if ((rgb255[i][mx] * 10) < (rgb255[i][mn] * 15)) {
      hue[i] = 1023;
    }
  }
  for (i = 0; i < NUM_PH; i++) {											//バッファセット
    sndW[i] = (word)hue[i];
    if ((sndW[i + 5] > 409) || (sndW[i + 5] < 102)) {
      sndW[i] = 1023;
    }
  }
}

//****	白補正動作  ****************************************
//------------------------------------------------
void adjust_White()											//白補正動作
{
  int i, j, k;

	k = 0;
	for (i = 0; i < NUM_PH; i++) {								//今の測定値をEEPに
  	i1 = seek_Max(&(adData[i][0]), 3);
  	for (j = 0; j < 3; j++) {
    	wK[i][j] = (adData[i][i1] * 16) / adData[i][j];
      EEPROM.write(k, lowByte(wK[i][j]));
      k++;
      EEPROM.write(k, highByte(wK[i][j]));
      k++;
    }
  }
  for (i = 0; i < NUM_PH; i++) {								//今の輝度をEEPに
  	EEPROM.write(k, lowByte(cData[i][0]));
    k++;
    EEPROM.write(k, highByte(cData[i][0]));
    k++;
	}
  digitalWrite(LED_R, LOW);
  digitalWrite(LED_G, LOW);
  digitalWrite(LED_B, LOW);
  delay(200);
  digitalWrite(LED_R, HIGH);
  digitalWrite(LED_G, LOW);
  digitalWrite(LED_B, LOW);
  delay(200);
  digitalWrite(LED_R, LOW);
  digitalWrite(LED_G, HIGH);
  digitalWrite(LED_B, LOW);
  delay(200);
  digitalWrite(LED_R, LOW);
  digitalWrite(LED_G, LOW);
  digitalWrite(LED_B, HIGH);
  delay(200);
}
//****	最大最小計算  **************************************
//------------------------------------------------
int seek_Max(word *d, int n)						//配列の中の最大値の添え字を探す
{
  int i;
  int ans=0;

  for (i = 1; i < n; i++) {
    if (*(d + ans) < * (d + i)) {
      ans = i;
    }
  }

  return ans;
}
//------------------------------------------------
int seek_Min(word *d, int n)						//配列の中の最小値の添え字を探す
{
  int i;
  int ans=0;

  for (i = 1; i < n; i++) {
    if (*(d + ans) > *(d + i)) {
      ans = i;
    }
  }

  return ans;
}

//****	EEPROM読み込み  ************************************
void Read_EEP()					// 補正値をEEPROMから読み出す
{
  int i, j, k;                          // ループのカウンタ
  byte wh, wl;

  k = 0;
  for (i = 0; i < NUM_PH; i++) {
    for (j = 0; j < 3; j++) {
      wl = EEPROM.read(k);
      k++;
      wh = EEPROM.read(k);
      k++;
      wK[i][j] = (word)wh << 8;
      wK[i][j] += (word)wl;
    }
  }
  for (i = 0; i < NUM_PH; i++) {
    wl = EEPROM.read(k);
    k++;
    wh = EEPROM.read(k);
    k++;
    cData0[i] = (word)wh << 8;
    cData0[i] += (word)wl;
  }
}

//****	I2C通信処理  ***************************************
//------------------------------------------------
void i2cReceiveEvent(int numBytes)			//データ受信時の処理
{
  int n = 0;
  int i;

  while (0 < Wire.available()) {
    rBuf[n] = (byte)Wire.read();
    n++;
  }
  command = rBuf[1];                // 受信データの2バイト目がコマンド
  switch (command) {
    case 0x80:
      numRep = 10;
      for (i = 0; i < 8; i++) {
        sBuf[i] = 0x01;
      }
      for (i = 8; i < 10; i++) {
        sBuf[i] = 0x01;
      }
      break;
    case 0x13:
      break;
    case 0x14:
      break;
    case 0x81:
      if (rBuf[2] == 0x0a) {
        numRep = 22;
        for (i = 0; i < 10; i++) {
          sBuf[i * 2] = lowByte(sndW[i]);
          sBuf[i * 2 + 1] = highByte(sndW[i]);
        }
        sBuf[20] = 0;
        sBuf[21] = 0;
      }
      else {
        numRep = 2;
        sBuf[0] = lowByte(sndW[rBuf[2]]);
        sBuf[1] = highByte(sndW[rBuf[2]]);
      }
      break;
    default:
      break;
  }
}
//------------------------------------------------
void i2cSendEvent()											//データ送信時の処理
{										
  Wire.write(sBuf, numRep);
}
//------------------------------------------------

